package ru.dbotthepony.mc.otm.block.entity.decorative

import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.MenuProvider
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.entity.IRedstoneControlled
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.block.entity.SynchronizedRedstoneControl
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.menu.decorative.HoloSignMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MBlocks

class HoloSignBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryBlockEntity(MBlockEntities.HOLO_SIGN, blockPos, blockState), MenuProvider, IRedstoneControlled {
	override val redstoneControl = SynchronizedRedstoneControl(synchronizer) { _, _ -> setChanged() }

	var text by synchronizer.string("", name = "text", setter = { value, access, remote ->
		setChanged()
		access.write(value)
	})

	var locked = false

	override fun createMenu(p_39954_: Int, p_39955_: Inventory, p_39956_: Player): AbstractContainerMenu {
		return HoloSignMenu(p_39954_, p_39955_, this)
	}

	override fun getDisplayName(): Component {
		return MBlocks.HOLO_SIGN.name
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt[TEXT_KEY] = text
		nbt[REDSTONE_CONTROL_KEY] = redstoneControl.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		locked = nbt.getBoolean(LOCKED_KEY)

		if (locked) {
			text = nbt.getString(TEXT_KEY)
		} else {
			text = truncate(nbt.getString(TEXT_KEY))
		}

		redstoneControl.deserializeNBT(nbt[REDSTONE_CONTROL_KEY] as? CompoundTag)
	}

	companion object {
		const val TEXT_KEY = "SignText"
		const val LOCKED_KEY = "Locked"
		const val REDSTONE_CONTROL_KEY = "RedstoneControl"
		const val DEFAULT_MAX_NEWLINES = 8
		const val DEFAULT_MAX_LINE_LENGTH = 15
		private val NEWLINES = Regex("\r?\n")

		fun truncate(input: String): String {
			val lines = input.split(NEWLINES)
			val result = ArrayList<String>(lines.size.coerceAtMost(DEFAULT_MAX_NEWLINES))

			for (i in 0 until lines.size.coerceAtMost(DEFAULT_MAX_NEWLINES)) {
				if (lines[i].length > DEFAULT_MAX_LINE_LENGTH) {
					result.add(lines[i].substring(0, DEFAULT_MAX_LINE_LENGTH))
				} else {
					result.add(lines[i])
				}
			}

			return result.joinToString("\n")
		}
	}
}
